from typing import Any, ClassVar, Protocol


def unflatten(dictionary: dict[Any, Any]) -> dict[Any, Any]:
    result_dict: dict[Any, Any] = dict()
    for key, value in dictionary.items():
        parts = key.split(".")
        d = result_dict
        for part in parts[:-1]:
            if part not in d:
                d[part] = dict()
            d = d[part]
        d[parts[-1]] = value
    return result_dict


class IsDataclass(Protocol):
    """
    Protocol to represent a dataclass.

    This protocol is used to check if an object is a dataclass by checking
    for the presence of the `__dataclass_fields__` attribute.
    """

    __dataclass_fields__: ClassVar[dict[Any, Any]]



"""
This module provides a function to create an argument parser for command line arguments.

The `create_parser` function takes in the name of a dataclass, which represents the arguments
that can be parsed from the command line. It creates an instance of `argparse.ArgumentParser`
and adds arguments based on the fields of the dataclass.

The function also allows specifying a YAML file that contains additional arguments, which can
be turned into the dataclass using the `dacite` library. These arguments can be overwritten
by specifying them directly on the command line.

The function returns an instance of `MyParser`, which is a custom wrapper around
`argparse.ArgumentParser`. This wrapper provides additional functionality for parsing
and accessing the command line arguments.

Example usage:
    # Create an argument parser for MyArgs dataclass
    parser = create_parser(MyArgs)

    # Parse command line arguments
    args = parser.parse_args()

    # Access the parsed arguments
    print(args.my_argument)
"""

import argparse
from dataclasses import Field, fields
from typing import Any, Type

from parsley_coco.parser import Parsley
from parsley_coco.utils import IsDataclass


def create_parsley[T_Dataclass: IsDataclass](
    args_dataclass_name: Type[T_Dataclass],
    should_parse_command_line_arguments: bool = True,
) -> Parsley[T_Dataclass]:
    """
    Create an argument parser for command line arguments.

    Args:
        args_class_name: The name of the dataclass representing the arguments.
        should_parse_command_line_arguments: Whether to parse command line arguments.

    Returns:
        An instance of MyParser, which is a custom wrapper around argparse.ArgumentParser.

    """
    parser: argparse.ArgumentParser = argparse.ArgumentParser()

    # one can specify a path to a yaml file containing parameters
    # that can be turned into the class named args_class_name (with dacite)
    parser.add_argument(
        "--config_file_name",
        type=str,
        default=None,
        help="path to a yaml file with arguments for the script",
    )

    # one can  specify parameters from the class named args_class_name
    # that will overwrite the ones in the yaml file
    field: Field[Any]
    for field in fields(args_dataclass_name):
        parser.add_argument(
            str("--" + field.name), type=str, default=None, help="to be written"
        )

    my_parser: Parsley[T_Dataclass] = Parsley(
        parser=parser,
        args_dataclass_name=args_dataclass_name,
        should_parse_command_line_arguments=should_parse_command_line_arguments,
    )

    return my_parser



from dataclasses import dataclass
from parsley_coco.recursive_dataclass_with_path_to_yaml import resolve_yaml_to_base


@dataclass
class C:
    x: int
    y: str


@dataclass
class A:
    a: int
    roo: C


@dataclass
class HOP:
    i: int
    b: C


@dataclass
class BIM:
    i: int
    hop: HOP


@dataclass
class C:
    x: int
    y: str


@dataclass
class A2:
    a: int
    roo: C
    bim: BIM


def test_resolve_dataclass_from_yaml():
    final_result = resolve_yaml_to_base("tests/yaml_files/test_a.yaml", A)
    print(final_result)


def test_resolve_dataclass_from_yaml_2():
    final_result = resolve_yaml_to_base("tests/yaml_files/test_a2.yaml", A2)
    print(final_result)


if __name__ == "__main__":
    test_resolve_dataclass_from_yaml()
    test_resolve_dataclass_from_yaml_2()



from dataclasses import dataclass, field

from parsley_coco import create_parsley, Parsley


@dataclass
class C:
    x: int = 4578
    y: str = "uio"


@dataclass
class A:
    a: int
    roo: C


@dataclass
class HOP:
    i: int = 789
    b: C = field(default_factory=C)


@dataclass
class BIM:
    i: int = 159
    hop: HOP = field(default_factory=HOP)


@dataclass
class C:
    x: int = 123
    y: str = "456"


@dataclass
class A2:
    a: int = 56
    roo: C = field(default_factory=C)
    bim: BIM = field(default_factory=BIM)


def test_parse_arguments():
    pars: Parsley[A2] = create_parsley(
        args_dataclass_name=A2, should_parse_command_line_arguments=False
    )
    a2 = pars.parse_arguments(extra_args={})
    print(a2)

    a2 = pars.parse_arguments(
        extra_args={"config_file_name": "tests/yaml_files/test_a2.yaml"}
    )
    print(a2)


if __name__ == "__main__":
    test_parse_arguments()



from parsley_coco import create_parsley

from dataclasses import dataclass


@dataclass
class TestDataClass:
    first_attribute: int = 0


def test_creation():
    parsley = create_parsley(
        should_parse_command_line_arguments=False, args_dataclass_name=TestDataClass
    )
    parsley.parse_arguments(extra_args={})



from dataclasses import fields, Field, make_dataclass, is_dataclass
from typing import Optional, List, Tuple, Any, Type, get_type_hints

import yaml
from dacite import from_dict
from dacite.types import is_instance


def make_dataclass_with_optional_paths_ii(cls: Type[Any]) -> Type[Any]:
    assert is_dataclass(cls), f"{cls} must be a dataclass"

    new_fields: List[Tuple[str, Any, Field[Any]]] = []
    hints = get_type_hints(cls)

    for field in fields(cls):
        field_type = hints[field.name]

        # Always add the original field (make optional if it's a dataclass)
        if is_dataclass(field_type):
            new_fields.append((field.name, Optional[field_type], field))
            new_fields.append((f"{field.name}_path_to_yaml_file", Optional[str], field))
        else:
            new_fields.append((field.name, field_type, field))

    new_cls_name = cls.__name__ + "_with_potential_path"
    return make_dataclass(new_cls_name, new_fields)


def make_dataclass_with_optional_paths(cls: Type[Any]) -> Type[Any]:
    assert is_dataclass(cls), f"{cls} must be a dataclass"

    new_fields: List[Tuple[str, Any]] = []
    hints = get_type_hints(cls)

    for f in fields(cls):
        field_type = hints[f.name]

        # If it's a dataclass, add both the field and its '_path_to_yaml_file' sibling
        if is_dataclass(field_type):
            new_fields.append((f.name, Optional[field_type]))
            new_fields.append((f"{f.name}_path_to_yaml_file", Optional[str]))
        else:
            new_fields.append((f.name, field_type))

    new_cls_name = cls.__name__ + "_with_potential_path"
    return make_dataclass(new_cls_name, new_fields)


def resolve_yaml_to_base(yaml_path: str, base_cls: Type[Any]) -> Any:
    extended_cls = make_dataclass_with_optional_paths(base_cls)

    try:
        with open(yaml_path, "r") as f:
            yaml_data = yaml.safe_load(f)
    except IOError:
        raise Exception("Could not read file:", yaml_path)

    extended_obj = from_dict(data_class=extended_cls, data=yaml_data)

    return resolve_extended_object(extended_obj, base_cls)


def resolve_extended_object(extended_obj: Any, base_cls: Type[Any]) -> Any:
    resolved_data = {}

    for field in fields(base_cls):
        base_field_type = field.type
        val = getattr(extended_obj, field.name, None)

        if is_dataclass(base_field_type):
            path_val = getattr(extended_obj, f"{field.name}_path_to_yaml_file", None)

            if path_val and not val:
                assert isinstance(base_field_type, type)
                resolved_val = resolve_yaml_to_base(path_val, base_field_type)
            elif val:
                assert isinstance(base_field_type, type)
                resolved_val = resolve_extended_object(val, base_field_type)
            else:
                resolved_val = None  # or raise, depending on strictness

            resolved_data[field.name] = resolved_val
        else:
            resolved_data[field.name] = val

    return from_dict(data_class=base_cls, data=resolved_data)
